/* Ioana Neamtu
 * PM 2017
 * Grupa 335CC
 */

/*
	l
	i
	n
	e
	s
	  0 1 2 3 4 5 6 7 8   0 1 2
 	  _ _ _ _ _ _ _ _ _   _ _ _
 	0|* * * * * * * * *   * * *
 	1|* * * * * * * * *   * * *   MiniBoard
 	2|* * * * * * * * *   * * *
 	3|* * * * * * * * *   * * *  ScoreTable
 	4|* * * * * * * * *    
 	5|* * * * * * * * *
 	6|* * * * * * * * *
 	7|* * * * * * * * *
 	8|* * * * * * * * *
 	 
 	11 10 9 8 7 6 5 4 3   2 1 0  = i
 		MegaBoard
 	Serializare
 */

#define F_CPU 16000000

#include <avr/io.h>
#include <util/delay.h>
//A port for columns
#define GREEN_CLOCK	PA0
#define GREEN_LATCH	PA1
#define GREEN_DATA	PA2

#define RED_CLOCK	PA5
#define RED_LATCH	PA6
#define RED_DATA	PA7

#define COLS_DDR 	DDRA
#define COLS_PORT	PORTA

//C port for lines
#define LINES_CLOCK	PC0
#define LINES_LATCH	PC1
#define LINES_DATA	PC2

#define LINES_DDR 	DDRC
#define LINES_PORT	PORTC

//B register for keyboard
#define KB_DDR DDRB
#define KB_PORT PORTB
#define KB_PIN PINB

#define LOST 3
#define NEUTRAL 0
#define GREEN_PLAYER 1
#define RED_PLAYER 2
#define ERROR 2

#define MEGAGAME_SIZE 9
#define MINIGAME_SIZE 3

#define endOfGame(score1, score2) ((score1 + score2 == 3) ? (1) : (0))

typedef enum {CHOOSE_BOARD, CHOOSE_MOVE, GAME_DONE, SET_DONE} GameState;
GameState game_state;


unsigned char keyPressed;
unsigned char currentPlayer;
int activeRow;
int activeCol;
int moveRow;
int moveCol;
int invalidMoveFlag = 0;      // Valid move

unsigned char MegaGame [MEGAGAME_SIZE][MEGAGAME_SIZE];
unsigned char MiniGame[MINIGAME_SIZE][MINIGAME_SIZE];
unsigned char MiniGameCopy[MINIGAME_SIZE][MINIGAME_SIZE];
unsigned char ScoreTable[MINIGAME_SIZE];
unsigned char ScoreTableCopy[MINIGAME_SIZE];
int score[MINIGAME_SIZE];	
int boardsWon[MINIGAME_SIZE + 1];	

// Hardware control functions
unsigned char get_key();
void uploadLine (int lineNumber);

// Game functions
void initPlayground(unsigned char player);
void newGame(unsigned char player); 

// Miniboard functions
void updateMiniboard(int r, int c, unsigned char currentPlayer);
void getEmptyMiniboards(unsigned char currentPlayer);
int checkMiniboard(int r, int c, unsigned char currentPlayer);

//Megaboard functions
void updateBoard(int r, int c, unsigned char currentPlayer);
int checkBoard(unsigned char currentPlayer);
void winner(unsigned char currentPlayer);

//Scoretable functions
void restoreScoreTable();
void updateScoreTable(unsigned char currentPlayer);

void invalidMove();

int main(void) {
	
	initPlayground(GREEN_PLAYER);
	COLS_DDR |= ( ( 1 << GREEN_CLOCK ) | ( 1 << GREEN_LATCH ) | ( 1 << GREEN_DATA ) | ( 1 << RED_CLOCK ) | ( 1 << RED_LATCH ) | ( 1 << RED_DATA ) ); //iesiri coloane
	COLS_PORT = 0b00000000;
	LINES_DDR |= ( ( 1 << LINES_CLOCK ) | ( 1 << LINES_LATCH ) | ( 1 << LINES_DATA )  ); //iesiri LINII
	LINES_PORT = 0b00000000;
	
	while(1){
		keyPressed = get_key();
		_delay_us(1000);
		if (keyPressed != 0xFF)
		{
			// Valid input key
			unsigned char r = (keyPressed >> 4);
			unsigned char c = (keyPressed % 16);
			if (invalidMoveFlag)
				restoreScoreTable();

			switch(game_state) {
				case CHOOSE_BOARD:
					if (r < 3 && c < 3) 
					{
						if (MiniGameCopy[r][c] == NEUTRAL)
						{
							// Valid move	
							// Set active miniboard						
							game_state = CHOOSE_MOVE;
							activeRow = r;
							activeCol = c;
							// Update miniboard
							updateMiniboard(r,c, currentPlayer);
						}
						else
						{
							invalidMove();
						}
					}
					else 
					{
							invalidMove();
					}						
					break;
				case CHOOSE_MOVE:
					if (r < 3 && c < 3)
					{
						// Get current's move row and column
						moveRow = activeRow * MINIGAME_SIZE + r;
						moveCol = activeCol * MINIGAME_SIZE + c;

						if (MegaGame[moveRow][moveCol] == NEUTRAL) {
							updateBoard(moveRow, moveCol, currentPlayer);
							if (checkMiniboard(activeRow, activeCol, currentPlayer) == 1) {
								int r = checkBoard(currentPlayer);
								if (r == 1 || r == -1) 
									break;
							}
							switch (currentPlayer) {
								case RED_PLAYER: 
									currentPlayer = GREEN_PLAYER;
									break;
								case GREEN_PLAYER:
									currentPlayer = RED_PLAYER;
									break;
							}

							if (MiniGameCopy[r][c] == NEUTRAL)
							{
								// Valid move
								game_state = CHOOSE_MOVE;
								activeRow = r;
								activeCol = c;									
								updateMiniboard(r, c, currentPlayer);
								break;
							}
							
							else {
								// Current move to a won miniboard
								// Next player can move in any empty board
								// He has to chose the board first
								game_state = CHOOSE_BOARD;
								getEmptyMiniboards(currentPlayer);
								break;
							}
																				
						}
						else 
						{
							invalidMove();
						}
					}
					else 
					{
						invalidMove();
					}
					break;
				case GAME_DONE:
					if ( r == 3 && c == 0) 
					{
						// new game button has been pressed
						newGame(currentPlayer);
					}
					else 
					{
						invalidMove();
					}
					break;	
				case SET_DONE:
					if ( r == 3 && c == 1) 
					{
						initPlayground(currentPlayer);
					}					
			}

		}
		_delay_us(500);
		// clock = 0
		LINES_PORT &= ~(1 << LINES_CLOCK);
		// latch = 0
		LINES_PORT &= ~(1 << LINES_LATCH);
		// load line 0
		uploadLine (0);
		
		
		// set the line activator 
		LINES_PORT |= (1 << LINES_DATA);
		//clock = 1
		LINES_PORT |= (1 << LINES_CLOCK);
		//latch = 1
		LINES_PORT |= (1 << LINES_LATCH);
		LINES_PORT &= ~(1 << LINES_DATA);

		// load the others line
		for ( int line = 1 ; line < 9; line++ )
		{
			_delay_us(500);
			LINES_PORT &= ~(1 << LINES_CLOCK);
			LINES_PORT &= ~(1 << LINES_LATCH);
			uploadLine(line);
			LINES_PORT |= (1 << LINES_CLOCK);
			LINES_PORT |= (1 << LINES_LATCH);
			
			
		}
		LINES_PORT &= ~(1 << LINES_DATA);
	}
	return 0;
}
void newGame(unsigned char player) 
{
	// Init the game table
	for (int i = 0; i  < MEGAGAME_SIZE; i++) 
	{
		for (int j = 0; j < MEGAGAME_SIZE; j++)
			MegaGame[i][j] = NEUTRAL;
	}
	// Init the auxiliar game board
	for (int i = 0; i < MINIGAME_SIZE; i++) 
	{
		for (int j = 0; j < MINIGAME_SIZE; j++) 
		{
			MiniGame[i][j] = player;
			MiniGameCopy[i][j] = NEUTRAL;
		}
	}

	game_state = CHOOSE_BOARD;
	currentPlayer = player;
	boardsWon[RED_PLAYER] = 0;
	boardsWon[GREEN_PLAYER] = 0;
	boardsWon[LOST] = 0;
}
void initPlayground (unsigned char player)
{
	
	newGame(player);
	for (int i = 0; i < MINIGAME_SIZE; i++) 
	{
		ScoreTable[i] = NEUTRAL;
		ScoreTableCopy[i] = NEUTRAL;
	}

	score[RED_PLAYER] = 0;
	score[GREEN_PLAYER] = 0;
	
}

void uploadLine ( int lineNumber )
{
	
	for (int i = 0; i < 12; i++)
	{
		// Set Red and Green clock to 0
		COLS_PORT &= ~(1 << GREEN_CLOCK);
		COLS_PORT &= ~(1 << GREEN_LATCH);
		COLS_PORT &= ~(1 << RED_CLOCK);
		COLS_PORT &= ~(1 << RED_LATCH);

		if ( i < 3 ){
			if  ( lineNumber < 3 ) 
			{
				// Print from auxiliar board as specifed before
				if ( MiniGame [ lineNumber ][ 2 - i ] == GREEN_PLAYER )
				{
					COLS_PORT |= (1 << GREEN_DATA);
				}

				if ( MiniGame [ lineNumber ][ 2 - i ] == RED_PLAYER )
				{
					COLS_PORT |= (1 << RED_DATA);
				}
			}
			if ( lineNumber == 3 ) 
			{
				// Line 3 includes ScoreTable
				if ( ScoreTable [ 2 - i ] == GREEN_PLAYER )
				{
					COLS_PORT |= (1 << GREEN_DATA);
				}

				if ( ScoreTable [ 2 - i ] == RED_PLAYER ) 
				{
					COLS_PORT |= (1 << RED_DATA);
				}
			}
		}
		if  ( i >= 3 )  
		{
			if ( MegaGame [ lineNumber ][ 11 - i ] == GREEN_PLAYER ) 
			{
					COLS_PORT |= (1 << GREEN_DATA);
			}
	
			if ( MegaGame [ lineNumber ][ 11 - i ] == RED_PLAYER ) 
			{					
					COLS_PORT |= (1 << RED_DATA);
			}
		}
		
		COLS_PORT |= (1 << GREEN_CLOCK);//clock = 1
		COLS_PORT &= ~(1 << GREEN_DATA);
		COLS_PORT |= (1 << RED_CLOCK);//clock = 1
		COLS_PORT &= ~(1 << RED_DATA);
	}

	COLS_PORT |= (1 << GREEN_LATCH);//latch = 1
	COLS_PORT |= (1 << RED_LATCH);//latch = 1
	
}

unsigned char get_key()
{
	
	// Set pins as inputs
	KB_DDR &= ~(0xff);
	// Set pull up 
	KB_PORT |= ((1 << 4) | (1 << 5) | (1 << 6) | (1 << 7));

	for( unsigned char c = 0 ; c < 4 ; c ++)
	{
		// Set a columns as output
		KB_DDR |= ( 0x08 >> c); 
		KB_PORT &= ~(0x08>>c); 

		for( unsigned char r = 0 ; r < 4 ; r ++)
		{
			if ( ( KB_PIN & (0x10 << r) ) == 0)
			{
				while ( ( KB_PIN & (0x10 << r) ) == 0)
				{;}
				_delay_ms(50);
				return r*16 + c ;
				
			}
			
		}

		KB_DDR &= ~(0X08 >> c);	
	}	

	return 0xFF;
}

void updateMiniboard(int r, int c, unsigned char currentPlayer) {
	// The player has chosen the game to play
	for (int i = 0; i < MINIGAME_SIZE; i++) {
		for (int j = 0; j < MINIGAME_SIZE; j++) {
			MiniGame[i][j] = NEUTRAL;
		}
	}
	MiniGame[r][c] = currentPlayer; 
}

void updateBoard(int r, int c, unsigned char currentPlayer) {
	MegaGame[r][c] = currentPlayer;
}

void getEmptyMiniboards(unsigned char currentPlayer) {
	for (int  i = 0; i < MINIGAME_SIZE; i++) 
		for (int j = 0; j < MINIGAME_SIZE; j++) 
			if (MiniGameCopy[i][j] == NEUTRAL)
				MiniGame[i][j] = currentPlayer;
			else MiniGame[i][j] = NEUTRAL;
}

int checkMiniboard(int r, int c, unsigned char currentPlayer) {

	int won = 0;
	// Check lines
	for (int i = r * MINIGAME_SIZE; i < r * MINIGAME_SIZE + MINIGAME_SIZE && won == 0; i++) {
		if (MegaGame[i][c*MINIGAME_SIZE] == currentPlayer &&
			MegaGame[i][c*MINIGAME_SIZE + 1] == currentPlayer &&
			MegaGame[i][c*MINIGAME_SIZE + 2] == currentPlayer)
				won = 1;
	}
	// Check columns
	for ( int  j = c * MINIGAME_SIZE; j < c * MINIGAME_SIZE + MINIGAME_SIZE && won == 0; j++) {
		if (MegaGame[r*MINIGAME_SIZE][j] == currentPlayer &&
			MegaGame[r*MINIGAME_SIZE + 1][j] == currentPlayer &&
			MegaGame[r*MINIGAME_SIZE + 2][j] == currentPlayer)
				won = 1;
	}

	// Check first diagonal
	if (won == 0 && 
		MegaGame[r * MINIGAME_SIZE][ c * MINIGAME_SIZE] == currentPlayer &&
		MegaGame[r * MINIGAME_SIZE + 1][ c * MINIGAME_SIZE + 1] == currentPlayer &&
		MegaGame[r * MINIGAME_SIZE + 2][ c * MINIGAME_SIZE + 2] == currentPlayer)
			won = 1;
	if (won == 0 &&
		MegaGame[r * MINIGAME_SIZE][ c * MINIGAME_SIZE + 2] == currentPlayer &&
		MegaGame[r * MINIGAME_SIZE + 1][ c * MINIGAME_SIZE + 1] == currentPlayer &&
		MegaGame[r * MINIGAME_SIZE + 2][ c * MINIGAME_SIZE + 0] == currentPlayer)
			won = 1;

	// Check if the miniboard is equal
	int equalFlag = 1;
	if ( won == 0) 
	{
		for (int i = r * MINIGAME_SIZE; i < r * MINIGAME_SIZE + MINIGAME_SIZE && equalFlag == 1; i++) 
		{
			for (int j = c * MINIGAME_SIZE; j < c * MINIGAME_SIZE + MINIGAME_SIZE; j++)
			{
				if (MegaGame[i][j] == NEUTRAL) 
					{
						equalFlag = 0;
						break;

					}
			}
		}
		if (equalFlag) 
		{
			boardsWon[LOST] ++;
			MiniGameCopy[r][c] = LOST;
		}
		
	}
	
	// The player won the board
	if (won == 1) {
		MiniGameCopy[r][c] = currentPlayer;
		for ( int i = r * MINIGAME_SIZE; i < r * MINIGAME_SIZE + MINIGAME_SIZE; i++)
		{
			for ( int  j = c * MINIGAME_SIZE; j < c * MINIGAME_SIZE + MINIGAME_SIZE; j++) 
			{
				updateBoard(i,j, currentPlayer);
			}
		}
		boardsWon[currentPlayer] ++;
	}

	return won;
}

int checkBoard(unsigned char currentPlayer) 
{
	int won = 0;
	// Check lines
	for (int i = 0; i < MINIGAME_SIZE && won == 0; i++) {
		if (MiniGameCopy[i][0] == currentPlayer && 
			MiniGameCopy[i][1] == currentPlayer &&
			MiniGameCopy[i][2] == currentPlayer)
			won = 1;
	}

	// Check cols
	for (int i = 0; i < MINIGAME_SIZE && won == 0; i++) {
		if (MiniGameCopy[0][i] == currentPlayer &&
			MiniGameCopy[1][i] == currentPlayer &&
			MiniGameCopy[2][i] == currentPlayer)
			won = 1;
	}

	if (won == 0 && MiniGameCopy[0][0] == currentPlayer &&
		MiniGameCopy[1][1] == currentPlayer &&
		MiniGameCopy[2][2] == currentPlayer)
		won = 1;

	if (won == 0 && MiniGameCopy[0][2] == currentPlayer &&
		MiniGameCopy[1][1] == currentPlayer &&
		MiniGameCopy[2][0] == currentPlayer)
		won = 1;
	if (won == 0 && boardsWon[RED_PLAYER] + boardsWon[GREEN_PLAYER] + boardsWon[LOST] == 9) 
	{
		// Equal
		game_state = GAME_DONE;
		for ( int  i = 0; i < MINIGAME_SIZE; i++)
			for (int j = 0; j < MINIGAME_SIZE; j++)
				MiniGame[i][j] = NEUTRAL;
		return -1;
	}

	if (won == 1) 
	{
		for (int i = 0; i < MEGAGAME_SIZE; i++) 
		{
			for(int j = 0; j < MEGAGAME_SIZE; j++) 
			{
				updateBoard(i,j, currentPlayer);
			}
		}

		for ( int  i = 0; i < MINIGAME_SIZE; i++)
			for (int j = 0; j < MINIGAME_SIZE; j++)
				MiniGame[i][j] = NEUTRAL;
		score[currentPlayer] ++;
		updateScoreTable(currentPlayer);
		if (score[currentPlayer] == 2) 
		{
			game_state = SET_DONE;
			winner(currentPlayer);
		}
		else
		{
			game_state = GAME_DONE;
		}
	}


	return won;
}

void invalidMove() 
{
	for (int i = 0; i < MINIGAME_SIZE; i++) {
		ScoreTableCopy[i] = ScoreTable[i];
		ScoreTable[i] = ERROR;
	}
	invalidMoveFlag = 1;
}
void restoreScoreTable() 
{
	for (int i = 0; i < MINIGAME_SIZE; i++)
		ScoreTable[i] = ScoreTableCopy[i];
}

void updateScoreTable(unsigned char currentPlayer) 
{
	for (int  i = 0; i < MINIGAME_SIZE; i++) 
	{
		if (ScoreTableCopy[i] == NEUTRAL)
		{
			ScoreTableCopy[i] = currentPlayer;
			ScoreTable[i] = currentPlayer;
			break;
		}
	}
}

void winner(unsigned char currentPlayer) 
{
	for (int i = 0; i < MEGAGAME_SIZE; i++)
	{
		for (int  j = 0; j < MEGAGAME_SIZE; j++)
		{
			if (j == 0 || j == MEGAGAME_SIZE - 1)
			{
				MegaGame[i][j] = currentPlayer;
			}
			else 
				MegaGame[i][j] = NEUTRAL;
		}
	}

	MegaGame[MEGAGAME_SIZE - 1][1] = currentPlayer;
	MegaGame[MEGAGAME_SIZE - 2][2] = currentPlayer;
	MegaGame[MEGAGAME_SIZE - 3][3] = currentPlayer;
	MegaGame[MEGAGAME_SIZE - 4][4] = currentPlayer;
	MegaGame[MEGAGAME_SIZE - 3][5] = currentPlayer;
	MegaGame[MEGAGAME_SIZE - 2][6] = currentPlayer;
	MegaGame[MEGAGAME_SIZE - 1][7] = currentPlayer;
}